unit Cybrix;

interface

{ Cybrix.Pas: version 1.53, 04-08-97, 17:34h

  SORRY ABOUT ALL THE DUTCH COMMENTS...


  Properties:
    DrawInner: Draw inner block y/n
    DrawBorder: Draw border y/n
    backColor: Background color
    Columns: No. of columns (1..27)
    Rows: No. of rows (3..60)
	
  Important methods:
    Newgame: Start a new game (clear grid etc)
    StopGame: Stops current game
    Movedown: Move active piece 1 down. Returns 0 if everything OK,
              -99 if game over (grid full), otherwise returns number
              of blocks deleted.

    MoveLeft: Move active piece 1 left
    MoveRight: ''    ''    ''   1 right
    SendToBottom: Send active piece to bottom grid
    redraw: redraws grid
    Create: constructor for the component

  Event handlers:
    OnGameOver:
      usefull for gameover-proc (overrides the -99-check in MoveDown())
    OnDeleteblock:
      usefull for maintaining the score and level-changing (overrides the
      number of blocks deleted from function MoveDown())

    Enjoy the game!


Arjan Scherpenisse
A.Scherpenisse@inter.NL.net

Rijksweg 204
6247 AN Rijckholt
the Netherlands
+31 43 4083518

Revision History:
27-08-97: Started with the component (v0.00)
28-08-97: First version finished  (v1.00)
          Made palette-icon (TCybrix.dcr)
30-08-97: * Added cool graphics (btRound, btButton) (v1.50)
          * Added two event handlers (v1.501)
31-08-97: * Fixed bug with active piece moving out of screen to the right
          * Fixed bug with complex blocks (CheckAndUpdate completely rewritten) (v1.51)
          ^^^ Thanks to Marco Bianchini! ^^^
01-08-97: Added 2 properties for controlling no. of cols and rows
          Started with background-bitmap instead of single color, doesn't work
          well (Sharing violation error at adress FFFFFF) (1.52)
04-08-97: Eliminated 2 compiler warnings and 4 hints (1.53)

************************************************************ 	
 
}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

const mxRows        = 20;
      mxColumns     = 9;
      mxPieceLength = 3;
      mxColors      = 4;

      clColor1      = clRed;
      clColor2      = clBlue;
      clColor3      = clWhite;
      clColor4      = clYellow;

      ldHorizontal  = 101;
      ldVertical    = 102;
      ldDiagright   = 103;
      ldDiagleft    = 104;

      GameOver      = -99;

type
  TBlockType = (btNormal, btRound, btButton);
  TbackType  = (btColor,btPicture);

  TBlock = record
    DoDelete,     
    Filled: Boolean;
    Color: TColor;
  End;

  TBlockGrid = array[0..27,0..60] of TBlock;

//  TBlockColors = (clRed,clBlue,clYellow,clWhite);
  TPiece  = class
  public
    Enabled: Boolean;
    X,Y:Integer;
    Colors: Array[0..mxPieceLength-1] of TColor;
    constructor Create;
    procedure SetPos(tx,ty: Integer);
    procedure SetColors(C1,C2,C3: TColor);
    procedure RandomColors;
    procedure CycleUp;
    procedure CycleDown;
  End;


  TCybrix = class(TGraphicControl)
  private
    FDrawBorder,
    FDrawInner:  Boolean;
    FBackColor:  TColor;
    Fgrid: TBlockgrid;
    FPlaying: Boolean;
    ActivePiece: TPiece;
    FBlockType: TBlockType;
    FColumns,
    FRows: Integer;
    FbackPicture: TPicture;
    FbackType: TbackType;

    FOnGameOver: TNotifyEvent;
    FOnDeleteBlock: TNotifyEvent;

    Procedure DrawPiece;
    Procedure HidePiece;
    Procedure DrawBlock(X,Y: Integer; C:TColor);
    Procedure FSetbackColor(C:TColor);
    Procedure FSetDrawInner(B:Boolean);
    Procedure FSetDrawBorder(B:Boolean);
    Procedure FSetBlockType(T:TblockType);
    Procedure FSetRows(N: Integer);
    Procedure FsetColumns(N:Integer);
    Procedure FSetbackPicture(P: TPicture);

    Procedure ClearGrid;
    Procedure DeleteLine(Y:Integer);
    Procedure ClearRow(Y:Integer);
    Procedure DeleteBlock(X,Y:Integer);
    Procedure SetBlockAt(X,Y: Integer; C:TColor);
    Procedure DeletemarkedBlocks;
    Function CheckAndUpdate: integer;
    Function CheckgameOver: Boolean;
  protected
    procedure Paint; override;
  public
    Procedure Rotate;
    Function MoveDown: Integer;
    Function MoveLeft: Integer;
    Function MoveRight: Integer;
    Procedure SendToBottom;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Redraw;
    procedure NewGame;
    procedure StopGame;
  published
    property DrawInner: Boolean read FDrawInner write FSetDrawInner default false;
    property DrawBorder: Boolean read FDrawBorder write FSetDrawBorder default true;
    property Backgroundcolor: TColor read FBackColor write FSetBackColor default clBlack;
    property BlockType: TBlockType read FBlockType write FSetBlockType default btNormal;
    property Columns: Integer read FColumns write FSetColumns default mxColumns;
    property Rows: Integer read FRows write FSetRows default mxRows;
    property BackGroundType: TbackType read FBackType write FBackType default btColor;
    property backPicture: TPicture read FBackPicture write FSetbackPicture;


    property OngameOver: TNotifyEvent read FOngameover write Fongameover;
    property OnDeleteBlock: TNotifyEvent read FOnDeleteblock write FOndeleteblock;
  end;

procedure Register;

implementation

Procedure TCybrix.FSetbackPicture(P: TPicture);
Begin
    FBackPicture:=P;
    BackGroundType:=btPicture;
    Paint;
End;

Procedure TCybrix.FSetRows(N: Integer);
begin
    if (N<=60) and (N>3) then
      begin
        FRows:=N;
        FSetColumns(FColumns);
        Paint;
      end
    else
      MessageDlg('Value out of range... [1,60]',mtWarning,[mbOK],0);
end;

Procedure TCybrix.FsetColumns(N:Integer);
Var A:Integer;
begin
  If (N>0) and (N<28) then
    begin
      FColumns:=N;
      If (csDesigning in Componentstate) then
      begin
        ClearGrid;
        For A:=0 to N-1 do
          If A mod 2=0 then
            SetBlockAt(A,FRows-1,clred)
          else
            SetBlockAt(A,FRows-2,clYellow);
       end;     
      Paint;
    end
  Else
      MessageDlg('Value out of range... [1,27]',mtWarning,[mbOK],0);
End;

Procedure TCybrix.FSetbackColor(C:TColor);
Begin
    backGroundType:=btColor;
    FbackColor:=C;
    Paint;
End;
Procedure TCybrix.FSetBlockType(T:TBlockType);
Begin
    FBlockType:=T;
    Paint;
End;

Procedure TCybrix.FSetDrawInner(B:Boolean);
Begin
    FdrawInner:=B;
    Paint;
End;

Procedure TCybrix.FSetDrawBorder(B:Boolean);
Begin
    FdrawBorder:=B;
    Paint;
End;


Procedure TCybrix.Paint;
Var A,B:Integer;
    R: TRect;
Begin
    With Canvas do
      begin
        If FBackType=btColor then
          begin
           Pen.Color:=FbackColor;
           For A:=0 to Height-1 do
             Begin
               MoveTo(0,A);
               LineTo(Width,A);
             End;
          End
        Else if FBackType=btPicture then
          begin
            R.Top:=0;
            R.left:=0;
            R.Right:=Width;
            R.Bottom:=Height;
            Canvas.StretchDraw(R,FbackPicture.Bitmap);
          end;

        For A:=0 to FColumns-1 do
          For B:=0 to FRows-1 do
            If FGrid[A,B].Filled then
              DrawBlock(A,B,FGrid[A,B].Color);
        If FPlaying then DrawPiece;      
       End;
End;

Procedure TCybrix.ClearGrid;
Var B:Integer;
Begin
      For B:=0 to FRows-1 do
        ClearRow(B);
End;


Procedure TCybrix.SetBlockAt(X,Y: Integer; C:TColor);
Begin
 //   While (FGrid[X,Y+1].Filled=false) and (Y+1<FRows) do
 //     Inc(Y);
    If C<>FBackColor then
     FGrid[X,Y].Filled:=True
    else
     FGrid[X,Y].Filled:=False;
    FGrid[X,Y].Color:=C;
    Fgrid[X,Y].DoDelete:=false;
End;

Procedure TCybrix.DeleteBlock(X,Y:Integer);
Var B:Integer;
Begin
    Fgrid[X,Y].DoDelete:=false; 
    For B:=Y downto 1 do
        Begin
          FGrid[X,B].Filled:=FGrid[X,B-1].Filled;
          FGrid[X,B].Color :=FGrid[X,B-1].Color;
        End;
    FGrid[X,0].Filled:=False;
    FGrid[X,0].Color:=clBlack;

    If Assigned(FOndeleteBlock) then
      FOndeleteBlock(Self);
End;

Procedure TCybrix.DeleteLine(Y:Integer);
Var A:Integer;
Begin
      For A:=0 to FColumns-1 do
        DeleteBlock(A,Y);
End;

Procedure TCybrix.ClearRow(Y:Integer);
Var A:Integer;
begin
    For A:=0 to FColumns-1 do
     With FGrid[A,Y] do
      Begin
        Filled:=false;
        Color:=clBlack;
        DoDelete:=false;
      End;
end;



constructor TCybrix.Create(AOwner: TComponent);
Begin
    Inherited Create(AOwner);

    ActivePiece:=TPiece.Create;
    ClearGrid;
    Randomize;
    FBackPicture:=TPicture.Create;
    If (csDesigning in ComponentState) then
      begin
        SetBlockAt(0,19, clYellow);
        SetBlockAt(1,18, clRed);
        SetBlockAt(2,19, clYellow);
        SetBlockAt(3,18, clRed);
        SetBlockAt(4,19, clYellow);
        SetBlockAt(5,18, clRed);
        SetBlockAt(6,19, clYellow);
        SetBlockAt(7,18, clRed);
        SetBlockAt(8,19, clYellow);
      end;
    FRows:=mxRows;
    FColumns:=mxColumns;
    FDrawBorder:=False;
    FDrawInner:=True;
    FBackColor:=clBlack;
    FBlockType:=btNormal;
    FBackType:=btColor;

    FPlaying:=false;
    ActivePiece.SetPos(FColumns div 2,0);
End;

Procedure TCybrix.redraw;
begin
    Paint;
end;

destructor TCybrix.Destroy;
Begin
    FBackPicture.Destroy;
    Inherited destroy;
End;

Function TCybrix.MoveDown: Integer;
Var A:Integer;
    Blocks: Integer;
Begin
   Blocks:=0;
   If FPlaying then
    If (FGrid[ActivePiece.X,ActivePiece.Y+mxPiecelength].Filled=false)
     and ((ActivePiece.Y+mxPieceLength)<FRows) then
      begin
        With ActivePiece do drawBlock(X,Y,FBackColor);
        Inc(ActivePiece.Y,1);
        DrawPiece;
      End
    else
      // Nieuw blokje
      Begin
        For A:=0 to mxPieceLength-1 do
          begin
            FGrid[ActivePiece.X,ActivePiece.Y+A].Filled:=true;
            FGrid[ActivePiece.X,ActivePiece.Y+A].Color:=ActivePiece.Colors[A];
          end;
        ActivePiece.SetPos(FColumns div 2,0);
        ActivePiece.RandomColors;
        Blocks:=CheckAndUpdate;
        If CheckgameOver then Blocks:=gameOver;
      End;
    MoveDown:=Blocks;  
End;

Function TCybrix.MoveLeft: Integer;
Var A:Integer;
    CanMove: Boolean;
Begin
   If FPlaying then
   begin
    CanMove:=True;
    For A:=0 to mxPieceLength-1 do
      If (FGrid[ActivePiece.X-1,ActivePiece.Y+A].Filled=True)
        or (ActivePiece.X-1<0) then canMove:=false;

    If canMove=true then
      Begin
        HidePiece;
        Dec(ActivePiece.X);
        DrawPiece;
      End;
    If canmove then moveleft:=1
    else moveleft:=0;
  End
  else
    Moveleft:=-1;
End;

Function TCybrix.MoveRight: Integer;
Var A:Integer;
    CanMove: Boolean;
Begin
   If FPlaying then
   begin
    CanMove:=True;
    For A:=0 to mxPieceLength-1 do
      If (FGrid[ActivePiece.X+1,ActivePiece.Y+A].Filled=True)
      or (ActivePiece.X+1>=FColumns)
        then canMove:=false;

      If canMove=true then
        Begin
          HidePiece;
          Inc(ActivePiece.X);
          DrawPiece;
        End;
    If canmove then moveright:=1
    else moveright:=0;
   end
  else
    Moveright:=-1;
End;

Procedure TCybrix.SendToBottom;
Var A: Integer;
Begin
    A:=ActivePiece.Y;
    HidePiece;
    While (Fgrid[ActivePiece.X,A+mxPiecelength].Filled=false)
      and (A+mxPiecelength<FRows) do Inc(A);
    ActivePiece.Y:=A;
    DrawPiece;
End;


Procedure TCybrix.DrawBlock(X,Y: Integer; C:TColor);
Var Xfact,Yfact: Single;
    X1,X2,Y1,Y2: Integer;

begin
    XFact:=Width / FColumns;
    YFact:=Height / FRows;

    X1:=Round(X*XFact);Y1:=Round(Y*YFact);
    X2:=Round((X+1)*XFact)-1; Y2:=Round((Y+1)*YFact)-1;
    Canvas.Pen.Color:=C;
    Canvas.Brush.Style:=bsSolid;
    If C=FBackColor then
        Begin
          canvas.Brush.Color:=FBackColor;
          Canvas.rectangle(X1,Y1,X2+1,Y2+1);
        End
    Else
      case FBlockType of
      btNormal:
       begin
        Canvas.Brush.Color:=FBackColor;
        If FDrawBorder then
          Canvas.rectangle(X1,Y1,X2,Y2);
        Canvas.Brush.Color:=C;
        If FDrawInner then
          Canvas.rectangle(X1+1,Y1+1,X2-1,Y2-1);
       end;
      btRound:
       begin
        canvas.Brush.Color:=FbackColor;
        If FDrawBorder then
          canvas.Ellipse(X1,Y1,X2+1,Y2+1);
        canvas.Brush.Color:=C;
        If FDrawInner then
          canvas.Ellipse(X1+1,Y1+1,X2,Y2);
      end;
    btButton:
      With canvas do
      If C<>FBackColor then
      Begin
        Brush.Color:=C;
        If FDrawBorder then
          Rectangle(X1+1,Y1+1,X2,Y2)
        else
          rectangle(X1+2,Y1+2,X2-1,Y2-1);
        Pen.Color:=clWhite;
        MoveTo(X1,Y2);
        LineTo(X1,Y1);LineTo(X2,Y1);
        Pen.Color:=clGray;
        MoveTo(X2,Y1);
        LineTo(X2,y2);LineTo(X1,Y2);
      end;
    end;
End;

Procedure TCybrix.DrawPiece;
var A:Integer;
Begin
    For A:=0 to mxPieceLength-1 do
      DrawBlock(ActivePiece.X, ActivePiece.Y+A, ActivePiece.Colors[A]);
End;

Procedure TCybrix.HidePiece;
var A:Integer;
Begin
    For A:=0 to mxPieceLength-1 do
      DrawBlock(ActivePiece.X, ActivePiece.Y+A, FBackColor);
End;

Procedure TCybrix.NewGame;
Begin
    clearGrid;
    FPlaying:=True;
    ActivePiece.Setpos(4,0);
    Paint;
End;

Procedure TCybrix.Stopgame;
Begin
    FPlaying:=false;
End;

Procedure TCybrix.Rotate;
begin
    ActivePiece.CycleUp;
    DrawPiece;
end;


Function TCybrix.CheckAndUpdate:integer;
Var X,Y:Integer;
    C2,
    Changed: Boolean;
    C: TColor;
    BlocksDeleted: Integer;
begin
  BlocksDeleted:=0;
  repeat
    C2:=false;
    Changed:=false;
    For X:=0 to FColumns-1 do
      For Y:=0 to FRows-1 do
       begin
        C:=Fgrid[X,Y].Color;
        If (FGrid[X,Y].Filled) then
         begin
           // Verticaal
           if  (FGrid[X+2,Y].Color=C)
           and (FGrid[X+1,Y].Color=C) then
             begin
               FGrid[X,Y].DoDelete:=true;
               FGrid[X+2,Y].DoDelete:=true;
               FGrid[X+1,Y].DoDelete:=true;
               Changed:=True;
               C2:=True;
               Inc(BlocksDeleted);
             end;
           // Horizontaal
           if  (FGrid[X,Y+2].Color=C)
           and (FGrid[X,Y+1].Color=C) then
             begin
               FGrid[X,Y].DoDelete:=true;
               FGrid[X,Y+2].DoDelete:=true;
               FGrid[X,Y+1].DoDelete:=true;
               C2:=True;
               Changed:=True;
               Inc(BlocksDeleted);
             end;
           // Diagonaal : \
           if  (FGrid[X+2,Y+2].Color=C)
           and (FGrid[X+1,Y+1].Color=C) then
             begin
               FGrid[X,Y].DoDelete:=true;
               FGrid[X+2,Y+2].DoDelete:=true;
               FGrid[X+1,Y+1].DoDelete:=true;
               C2:=True;
               Changed:=True;
               Inc(BlocksDeleted);
             end;
           // Diagonaal : /
           if  (FGrid[X+2,Y-2].Color=C)
           and (FGrid[X+1,Y-1].Color=C) then
             begin
               C2:=True;
               FGrid[X,Y].DoDelete:=true;
               FGrid[X+2,Y-2].DoDelete:=true;
               FGrid[X+1,Y-1].DoDelete:=true;
               Changed:=True;
               Inc(BlocksDeleted);
             end;
          End;
       End;
        If Changed then
          begin
             DeleteMarkedBlocks;
             paint;
          End;
   Until C2=false;       
   CheckAndUpdate:=BlocksDeleted;
end;


Function TCybrix.CheckgameOver: Boolean;
Var X,Y,A:integer;
    T:Boolean;
begin
    T:=false;
    For A:=0 to FColumns-1 do
      If FGrid[A,0].Filled then
          T:=true;
    If T=true then
      begin
        For X:=0 to FColumns-1 do
          For Y:=0 to FRows-1 do
            SetBlockAt(X,Y,clColor1);
        paint;
        If Assigned(OngameOver) then
          FOngameOver(Self);
      End;
    CheckgameOver:=T;
End;

// TPiece class

constructor TPiece.Create;
Begin
    X:=5;
    Y:=0;
    RandomColors;
End;

procedure TPiece.SetColors(C1,C2,C3: TColor);
Begin
    Colors[0]:=C1;
    Colors[1]:=C2;
    Colors[2]:=C3;
End;

procedure TPiece.RandomColors;
Var A:integer;
Begin
    For A:=0 to mxPieceLength-1 do
       case Random(4)+1 of
       1: Colors[A]:=clColor1;
       2: Colors[A]:=clColor2;
       3: Colors[A]:=clColor3;
       4: Colors[A]:=clColor4;
       end;
end;

procedure TPiece.CycleUp;
Var C:TColor;
Begin
    C:=Colors[0];
    Colors[0]:=Colors[1];
    Colors[1]:=Colors[2];
    Colors[2]:=C; 
End;

procedure TPiece.CycleDown;
Var C:TColor;
Begin
    C:=Colors[2];
    Colors[2]:=Colors[1];
    Colors[1]:=Colors[0];
    Colors[0]:=C;
End;

procedure TPiece.SetPos(tx,ty: Integer);
Begin
    X:=tx;
    Y:=ty;
End;

Procedure TCybrix.DeletemarkedBlocks;
Var X,Y:Integer;
begin
    For X:=0 to FColumns-1 do
      For Y:=0 to FRows-1 do
        If Fgrid[X,Y].DoDelete then
          DeleteBlock(X,Y);
End;

// registratie in palet:

procedure Register;
begin
  RegisterComponents('Samples', [TCybrix]);
end;


end.
